#include "stdafx.h"
#include "CGenAlg.h"




//-----------------------------------constructor-------------------------
//
//	sets up the population with random floats
//
//-----------------------------------------------------------------------
CGenAlg::CGenAlg(int	  popsize,
                 double	MutRat,
                 double	CrossRat,
                 int	  numweights) :	m_iPopSize(popsize),
                                      m_dMutationRate(MutRat),
										                  m_dCrossoverRate(CrossRat),
										                  m_iChromoLength(numweights),
										                  m_dTotalFitness(0),
										                  m_cGeneration(0),
										                  m_iFittestGenome(0),
										                  m_dBestFitness(0),
										                  m_dWorstFitness(99999999),
										                  m_dAverageFitness(0)
{
	//initialise population with chromosomes consisting of random
	//weights and all fitnesses set to zero
	for (int i=0; i<m_iPopSize; ++i)
	{
		m_vecPop.push_back(SGenome());

		for (int j=0; j<m_iChromoLength; ++j)
		{
			m_vecPop[i].vecWeights.push_back(RandomClamped());
		}
	}
}


//---------------------------------Mutate--------------------------------
//
//	mutates a chromosome by perturbing its weights by an amount not 
//	greater than CParams::dMaxPerturbation
//-----------------------------------------------------------------------
void CGenAlg::Mutate(vector<double> &chromo)
{
	//traverse the chromosome and mutate each weight dependent
	//on the mutation rate
	for (unsigned int i=0; i<chromo.size(); ++i)
	{
		//do we perturb this weight?
		if (RandFloat() < m_dMutationRate)
		{
			//add or subtract a small value to the weight
			chromo[i] += (RandomClamped() * CParams::dMaxPerturbation);
		}
	}
}

//----------------------------------GetChromoRoulette()------------------
//
//	returns a chromo based on roulette wheel sampling
//
//-----------------------------------------------------------------------
SGenome CGenAlg::GetChromoRoulette()
{
 // generate a random number between 0 & total fitness count
 double Slice = (double)(RandFloat() * m_dTotalFitness);

 // this will be set to the chosen chromosome
 SGenome TheChosenOne = m_vecPop[0];
	
 // go through the chromosones adding up the fitness so far
 double FitnessSoFar = 0;
	
 for(int i = 0; i < m_iPopSize; ++i)
 {
	 FitnessSoFar += m_vecPop[i].dFitness;
		
	 // if the fitness so far > random number return the chromo at this point
	 if(FitnessSoFar >= Slice)
	 {
		TheChosenOne = m_vecPop[i];
//      m_vecChosen[i]++;
        break;
 	 }
 }

 return TheChosenOne;
}
	
//-------------------------------------Crossover()-----------------------
//	
//  given parents and storage for the offspring this method performs
//	crossover according to the GAs crossover rate
//-----------------------------------------------------------------------
void CGenAlg::Crossover(const vector<double> &mum,
                        const vector<double> &dad,
                        vector<double>       &baby1,
                        vector<double>       &baby2)
{
	// just return parents as offspring dependent on the rate
	// or if parents are the same
	if ( (RandFloat() > m_dCrossoverRate) || (mum == dad)) 
	{
		baby1 = mum;
		baby2 = dad;

		return;
	}

	// determine a crossover point
	int cp = RandInt(0, m_iChromoLength - 1);
    int i;

	// create the offspring
	for(i = 0; i < cp; ++i)
	{
		baby1.push_back(mum[i]);
		baby2.push_back(dad[i]);
	}

	for(; i < (int)mum.size(); ++i)
	{
		baby1.push_back(dad[i]);
		baby2.push_back(mum[i]);
	}
	
	
	return;
}

//-----------------------------------Epoch()-----------------------------
//
//	takes a population of chromosones and runs the algorithm through one
//	 cycle.
//	Returns a new population of chromosones.
//
//-----------------------------------------------------------------------
/*
vector<SGenome> CGenAlg::Epoch(vector<SGenome> &old_pop)
{
 // assign the given population to the classes population
 m_vecPop = old_pop;

 // reset the appropriate variables
 // Reset();

 // *Note: original code did a sort, then CalculateBestWorstAvTot()
 // However, after sorting, best were ordered to the end, 
 // thus losing knowledge of which genome won
 // (and which performed worst - though that info is not used at this time)
 // Also, after a sort we know the best and worst fitness without further effort.
 // Anyhow, since I want to know if a non-elite won,
 // I will reverse the order of these function calls...

 // calculate best, worst, average and total fitness
 CalculateBestWorstAvTot();

 // sort the population (for scaling and elitism)
 // sorted worst to best order
 sort(m_vecPop.begin(), m_vecPop.end());
*/

// Expects m_pGA->SortAndRank() was previously done on m_vecPop
vector<SGenome> CGenAlg::Epoch(void)
{

 // create a temporary vector to store new chromosones
 vector <SGenome> vecNewPop;

 // Now to add a little elitism we shall add in some copies of the
 // fittest genomes. Make sure we add an EVEN number or the roulette
 // wheel sampling will crash
 if(!(CParams::iNumCopiesElite * CParams::iNumElite % 2))
 {
	 GrabNBest(CParams::iNumElite, CParams::iNumCopiesElite, vecNewPop);
 }

 // now we enter the GA loop
 // repeat until a new population is generated
 while(vecNewPop.size() < (unsigned)m_iPopSize)
 {
	   // grab two chromosones
	   SGenome mum = GetChromoRoulette();
	   SGenome dad = GetChromoRoulette();

	   // create some offspring via crossover
	   vector<double> baby1, baby2;
	   Crossover(mum.vecWeights, dad.vecWeights, baby1, baby2);

	   // now we mutate
	   Mutate(baby1);
	   Mutate(baby2);

	   // now copy into vecNewPop population
	   vecNewPop.push_back(SGenome(baby1, 0));
	   vecNewPop.push_back(SGenome(baby2, 0));
 }

 // finished so assign new pop back into m_vecPop
 m_vecPop = vecNewPop;

 return m_vecPop;
}


//-------------------------GrabNBest----------------------------------
//
//	This works like an advanced form of elitism by inserting NumCopies
//  copies of the NBest most fittest genomes into a population vector
//--------------------------------------------------------------------

void CGenAlg::GrabNBest(int  NBest, const int NumCopies, vector<SGenome> &Pop)
{
 // Add the required amount of copies of the n most fittest to the supplied vector
 // Geo: These end up in worst to best order. Was that the intention?
 // That has no effect on anything.
 while(NBest--)
 {
	   Pop.push_back(m_vecPop[(m_iPopSize - 1) - NBest]);
 }
}

//-----------------------CalculateBestWorstAvTot-----------------------	
//
//	calculates the fittest and weakest genome and the average/total 
//	fitness scores
//---------------------------------------------------------------------
void CGenAlg::CalculateBestWorstAvTot(void)
{
 Reset();
	
 for(int i = 0; i < m_iPopSize; ++i)
 {
	 // Find the best...
	 if(m_vecPop[i].dFitness > m_dBestFitness)
	 {
		m_dBestFitness = m_vecPop[i].dFitness;
		m_iFittestGenome = i;
	 }
		
	 // Find the worst...
	 if(m_vecPop[i].dFitness < m_dWorstFitness)
	 {
		m_dWorstFitness = m_vecPop[i].dFitness;
		m_iWorstGenome = i;
	 }
		
	 m_dTotalFitness += m_vecPop[i].dFitness;
 }

 m_dAverageFitness = m_dTotalFitness / m_iPopSize;
}

void CGenAlg::SortAndRank(vector<SGenome> &vecPop)
{
 // Assign the given population to the classes population
 m_vecPop = vecPop;

 // *Note: original code did a sort, then CalculateBestWorstAvTot()
 // However, after sorting, best were ordered to the end, 
 // thus losing knowledge of which genome won
 // (and which performed worst - though that info is not used at this time)
 // Also, after a sort we know the best and worst fitness without further effort.
 // Anyhow, since I want to know if a non-elite won,
 // I will reverse the order of these function calls...

 // calculate best, worst, average and total fitness
 CalculateBestWorstAvTot();

 // sort the population (for scaling and elitism)
 // sorted worst to best order
 sort(m_vecPop.begin(), m_vecPop.end());
}

// Expects a population that was ranked & sorted by 
// SortAndRank(vector<SGenome> &vecPop) above.
// Sort them to resemble the output from Epoch()
// This makes no changes to local (to this class) var m_vecPop
vector<SGenome> CGenAlg::OrderForExport(vector<SGenome> &vecPop)
{
 // create a temporary vector to store new chromosones
 vector <SGenome> vecOrderedPop;

 if(!(CParams::iNumCopiesElite * CParams::iNumElite % 2))
 {
	 GrabNBest(CParams::iNumElite, CParams::iNumCopiesElite, vecOrderedPop);
 }
 // They are ordered worst to best,
 // and we just copied the best off the end of the array to the front of the new array.
 // Now we copy the rest over from the old array, stopping short before the remainder we already have...
 int i = 0;
 while(vecOrderedPop.size() < (unsigned)m_iPopSize)
 {
	   vecOrderedPop.push_back(vecPop[i++]);
 }

 return vecOrderedPop;
}


//-------------------------Reset()------------------------------
//
//	resets all the relevant variables ready for a new generation
//--------------------------------------------------------------
void CGenAlg::Reset()
{
	m_dTotalFitness		= 0;
	m_dBestFitness		= 0;
	m_dWorstFitness		= 9999999;
	m_dAverageFitness	= 0;
}

